import 'package:json_api/http.dart';
import 'package:json_api/src/client/disposable_handler.dart';
import 'package:json_api/src/client/persistent_handler.dart';
import 'package:json_api/src/client/request.dart';
import 'package:json_api/src/client/response.dart';

/// A basic JSON:API client.
///
/// The JSON:API [Request] is converted to [HttpRequest] and sent downstream
/// using the [handler]. Received [HttpResponse] is then converted back to
/// JSON:API [Response].
///
/// By default, the [DisposableHandler] is used which internally creates
/// a new instance of Dart built-in HTTP client for each request and then
/// disposes it. If you want more control of the underlying http client,
/// one option can be to use the [PersistentHandler]. To use another HTTP client,
/// such as [dio](https://pub.dev/packages/dio) implement your own wrapper.
///
/// The [codec] performs JSON encoding/decoding. The default implementation
/// uses native `dart:convert`. Provide your own [PayloadCodec] if you need
/// fine-grained control over JSON conversion.
class Client {
  const Client(
      {PayloadCodec codec = const PayloadCodec(),
      HttpHandler handler = const DisposableHandler()})
      : _codec = codec,
        _http = handler;

  final HttpHandler _http;
  final PayloadCodec _codec;

  /// Sends the [request] to the given [uri].
  Future<Response> send(Uri uri, Request request) async {
    final body = await _encode(request.document);
    final response = await _http.handle(HttpRequest(
        request.method,
        request.query.isEmpty
            ? uri
            : uri.replace(queryParameters: request.query),
        body: body)
      ..headers.addAll({
        'Accept': mediaType,
        if (body.isNotEmpty) 'Content-Type': mediaType,
        ...request.headers
      }));

    final document = await _decode(response);
    return Response(response, document);
  }

  Future<String> _encode(Object? doc) async =>
      doc == null ? '' : await _codec.encode(doc);

  Future<Map?> _decode(HttpResponse response) async =>
      response.hasDocument ? await _codec.decode(response.body) : null;
}
